/*****************************************************************************
 * loadmap.cpp - map class, the functions to load the map goes here
 ****************************************************************************/
#include <string>
#include "SDL.h"
#include "ccmap.h"
#include "common.h"
#include "compression.h"
#include "config.h"
#include "fcnc_endian.h"
#include "freecnc.h"
#include "game.h"
#include "inifile.h"
#include "logger.h"
#include "netconnection.h"
#include "playerpool.h"
#include "shpimage.h"
#include "unitandstructurepool.h"
#include "vfs.h"

/** Loads the maps ini file containing info on dimensions, units, trees
 * and so on.
 */

void CnCMap::loadIni()
{
    ConfigType config;
    INIFile *inifile;
    char *tmpname = new char[strlen(missionData.mapname)+5];
    config = getConfig();
    strcpy(tmpname, missionData.mapname);
    strcat(tmpname, ".INI");

    // Load the INIFile
    try {
        inifile = new INIFile(tmpname);
    } catch (int) {
        logger->error("Map \"%s\" not found.  Check your installation.\n",tmpname);
        throw LoadMapError();
    }
    delete[] tmpname;

    p::ppool = new PlayerPool(inifile, gamemode);

    if (inifile->readInt("basic", "newiniformat", 0) != 0) {
        logger->error("Red Alert maps not fully supported yet\n");
    }

    simpleSections(inifile);

    try {
        p::uspool = new UnitAndStructurePool();
    } catch (int) {
        throw LoadMapError();
    }

    if (gamemode == 0) {
        p::ppool->setLPlayer(missionData.player);
    }
    terraintypes = new Uint8[width*height];
    memset(terraintypes, 0, width*height);

    advancedSections(inifile);

    if (maptype == GAME_RA)
        unMapPack(inifile);

    p::uspool->preloadUnitAndStructures(missionData.buildlevel);
    // spawn player starts.
    if (gamemode > 0) {
        tmpname = new char[8];
        for (Uint8 i=0;i<config.totalplayers;++i) {
            sprintf(tmpname,"multi%i",i+1);
            p::ppool->getPlayerNum(tmpname);
            if ((i+1) == config.playernum) {
                p::ppool->setLPlayer(config.playernum,config.nick,config.side_colour,config.mside);
            }
        }
        delete[] tmpname;
        p::ppool->placeMultiUnits();
    }
    try {
        pips = new SHPImage("hpips.shp",mapscaleq);
    } catch(ImageNotFound&) {
        try {
            pips = new SHPImage("pips.shp", mapscaleq);
        } catch(ImageNotFound&) {
            logger->error("Unable to load the pips graphics!\n");
            throw LoadMapError();
        }
    }
    pipsnum = pc::imagepool->size()<<16;
    pc::imagepool->push_back(pips);
    if (maptype == GAME_RA) {
        try {
            char moveflsh[13] = "moveflsh.";
            strncat( moveflsh, missionData.theater, 3 );
            moveflash = new SHPImage(moveflsh,mapscaleq);
        } catch (ImageNotFound&) {
            logger->error("Unable to load the movement acknowledgement pulse graphic\n");
            throw LoadMapError();
        }
    } else {
        try {
            moveflash = new SHPImage("moveflsh.shp",mapscaleq);
        } catch (ImageNotFound&) {
            logger->error("Unable to load the movement acknowledgement pulse graphic\n");
            throw LoadMapError();
        }
    }
    flashnum = pc::imagepool->size()<<16;
    pc::imagepool->push_back(moveflash);
}


/** Function to load all vars in the simple sections of the inifile
 * @param pointer to the inifile
 */
void CnCMap::simpleSections(INIFile* inifile) {
    Uint32 temp;

    // The strings in the basic section
    missionData.brief = inifile->readString("BASIC", "BRIEF");
    missionData.action = inifile->readString("BASIC", "ACTION");
    missionData.player = inifile->readString("BASIC", "PLAYER");
    missionData.theme = inifile->readString("BASIC", "THEME");
    missionData.winmov = inifile->readString("BASIC", "WIN");
    missionData.losemov = inifile->readString("BASIC", "LOSE");
    missionData.buildlevel = inifile->readInt("BASIC", "BUILDLEVEL",1);

    // The ints (and one string) in the map section
    temp = inifile->readInt("MAP", "HEIGHT");
    if (temp == INIERROR) {
        logger->error("Error loading map: was unable to find height\n");
        throw LoadMapError();
    }
    height = temp;

    temp = inifile->readInt("MAP", "WIDTH");
    if (temp == INIERROR) {
        logger->error("Error loading map: was unable to find width\n");
        throw LoadMapError();
    }
    width = temp;

    temp = inifile->readInt("MAP", "X");
    if (temp == INIERROR) {
        logger->error("Error loading map: was unable to find \"x\"\n");
        throw LoadMapError();
    }
    x = temp;

    temp = inifile->readInt("MAP", "Y");
    if (temp == INIERROR) {
        logger->error("Error loading map: was unable to find \"y\"\n");
        throw LoadMapError();
    }
    y = temp;

    missionData.theater = inifile->readString("MAP", "THEATER");
    if (missionData.theater == NULL) {
        logger->error("Error loading map: missing theater\n");
        throw LoadMapError();
    }

    if (missionData.brief  == NULL || missionData.action  == NULL ||
        missionData.player == NULL || missionData.losemov == NULL ||
        missionData.theme  == NULL || missionData.winmov  == NULL) {
        logger->error("Error loading map: missing film data\n");
        throw LoadMapError();
    }
}

/** Function to load all the advanced sections in the inifile.
 * @param a pointer to the inifile
 */
void CnCMap::advancedSections(INIFile *inifile) {
    Uint32 keynum;
    int i;
    //char *line;
    char shpname[128];
    char trigger[128];
    char action[128];
    char type[128];
    char owner[128];
    int facing, health, subpos;
    int linenum, smudgenum, tmpval;
    Uint16 tx, ty, xsize, ysize, tmp2;
    std::map<std::string, Uint32> imagelist;
    std::map<std::string, Uint32>::iterator imgpos;
    SHPImage *image;
    TerrainEntry tmpterrain;
    INIKey key;

    Uint16 xwalk, ywalk, ttype;

    try {
        for( keynum = 0; ; keynum++ ) {
            key = inifile->readKeyValue("WAYPOINTS", keynum);
            if(sscanf(key->first.c_str(), "%d", &tmpval) == 1) {
                if( tmpval == 26 ) { /* waypoint 26 is the startpos of the map */
                    tmp2 = (Uint16)atoi(key->second.c_str());
                    //waypoints.push_back(tmp2);
                    tx = tmp2%64;
                    ty = tmp2/64;
                    scrollbookmarks[0].x = tx-x;
                    scrollbookmarks[0].y = ty-y;
                }
                if( tmpval < 8 ) {
                    tmp2 = (Uint16)atoi(key->second.c_str());
                    waypoints.push_back(tmp2);
                }
            }
        }
    } catch(int) {}
    p::ppool->setWaypoints(waypoints);

    INIFile *arts = new INIFile("art.ini");

    /* load the shadowimages */
    try {
        image = new SHPImage("shadow.shp", mapscaleq);
        numShadowImg = image->getNumImg();
        shadowimages = new SDL_Surface*[numShadowImg];
        for( i = 0; i < 48; i++ ) {
            image->getImageAsAlpha(i, &shadowimages[i]);
        }
        delete image;
    } catch(ImageNotFound&) {
        logger->warning("Unable to load \"shadow.shp\"\n");
        numShadowImg = 0;
    }
    /* load the smudge marks and the tiberium to the imagepool */
    if( maptype == GAME_TD ) {
        strncpy(shpname,"TI1.",128);
        strncat(shpname,missionData.theater,3);
        try { // TI1.SNO does not exist, since snow isn't in TD.
            image = new SHPImage(shpname, mapscaleq);
            pc::imagepool->push_back(image);
        } catch (ImageNotFound&) {
            logger->error("Could not load %s\n",shpname);
            throw LoadMapError();
        }
    } else if( maptype == GAME_RA ) {
        strcpy(shpname, "GOLD01."); // Red Alert Ore
        strncat(shpname, missionData.theater, 3);
        // No minerals on interior maps.
        if (strcasecmp("INT",missionData.theater)!=0) {
            try {
                image = new SHPImage(shpname, mapscaleq);
                pc::imagepool->push_back(image);
            } catch (ImageNotFound&) {
                logger->error("Could not load %s\n",shpname);
                throw LoadMapError();
            }
        } else {
            pc::imagepool->push_back(0);
        }
    } else {
        logger->error("Unsuported maptype\n");
        throw LoadMapError();
    }

    for( i = 1; i <= 6; i++ ) {
        sprintf(shpname, "SC%d.", i);
        strncat(shpname, missionData.theater, 3);
        image = new SHPImage(shpname, mapscaleq);
        pc::imagepool->push_back(image);
    }
    for( i = 1; i <= 6; i++ ) {
        sprintf(shpname, "CR%d.", i);
        strncat(shpname, missionData.theater, 3);
        image = new SHPImage(shpname, mapscaleq);
        pc::imagepool->push_back(image);
    }

    overlaymatrix = new Uint16[width*height];
    memset(overlaymatrix, 0, sizeof(Uint16)*width*height);

    try {
        for( keynum = 0; ;keynum++ ) {
            bool bad = false;
            key = inifile->readKeyValue("TERRAIN", keynum);
            /* , is the char which separate terraintype from action. */
            if( sscanf(key->first.c_str(), "%d", &linenum) == 1 &&
                    sscanf(key->second.c_str(), "%[^,],%s", shpname, trigger) >= 1 ) {
                /* Set the next entry in the terrain vector to the correct values.
                 * the map-array and shp files vill be set later */
                translateCoord(linenum, &tx, &ty);

                if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                    continue;
                }

                if( shpname[0] == 't' || shpname[0] == 'T' )
                    ttype = t_tree;
                else if( shpname[0] == 'r' || shpname[0] == 'R' )
                    ttype = t_rock;
                else
                    ttype = t_other_nonpass;

                /* calculate the new pos based on size and blocked */
                xsize = arts->readInt(shpname, "xsize",1);
                ysize = arts->readInt(shpname, "ysize",1);

                for( ywalk = 0; ywalk < ysize && ywalk + ty < height+y; ywalk++ ) {
                    for( xwalk = 0; xwalk < xsize && xwalk + tx < width + x; xwalk++ ) {
                        sprintf(type, "notblocked%d", ywalk*xsize+xwalk);
                        if( arts->readInt(shpname, type) == INIERROR ) {
                            terraintypes[(ywalk+ty-y)*width+xwalk+tx-x] = ttype;
                        }
                    }
                }

                linenum = xsize*ysize;
                do {
                    if (linenum == 0) {
                        logger->error("BUG: Couldn't find an entry in art.ini for %s\n",shpname);
                        bad = true;
                        break;
                    }
                    linenum--;
                    sprintf(type, "notblocked%d", linenum);
                } while(arts->readInt(shpname, type) == INIERROR);
                if (bad)
                    continue;

                tmpterrain.xoffset = -(linenum%ysize)*24;
                tmpterrain.yoffset = -(linenum/ysize)*24;

                tx += linenum%ysize;
                if( tx >= width+x ) {
                    tmpterrain.xoffset += 1+tx-(width+x);
                    tx = width+x-1;
                }
                ty += linenum/ysize;
                if( ty >= height+y ) {
                    tmpterrain.yoffset += 1+ty-(height+y);
                    ty = height+y-1;
                }

                linenum = (ty-y)*width + tx - x;
                strcat(shpname, ".");
                strncat(shpname, missionData.theater, 3);

                /* search the map for the image */
                imgpos = imagelist.find(shpname);

                /* set up the overlay matrix and load some shps */
                if( imgpos != imagelist.end() ) {
                    /* this tile already has a number */
                    overlaymatrix[linenum] |= HAS_TERRAIN;
                    tmpterrain.shpnum = imgpos->second << 16;
                    terrains[linenum] = tmpterrain;
                } else {
                    /* a new tile */
                    imagelist[shpname] = pc::imagepool->size();
                    overlaymatrix[linenum] |= HAS_TERRAIN;
                    tmpterrain.shpnum = pc::imagepool->size()<<16;
                    terrains[linenum] = tmpterrain;
                    image = new SHPImage(shpname, mapscaleq);
                    pc::imagepool->push_back(image);
                }
            }
        }
    } catch(int) {}

    try {
        for( keynum = 0;;keynum++ ) {
            key = inifile->readKeyValue("OVERLAY", keynum);
            if( sscanf(key->first.c_str(), "%d", &linenum) == 1 ) {
                strcpy(shpname, key->second.c_str());
                translateCoord(linenum, &tx, &ty);
                if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                    continue;
                }
                linenum = (ty-y)*width + tx - x;

                /* Set the next entry in the overlay vector to the correct values.
                 * the map-array and shp files vill be set later */
                if( !strcmp(shpname, "BRIK") || !strcmp(shpname, "SBAG") ||
                        !strcmp(shpname, "BARB") || !strcmp(shpname, "WOOD") ||
                        !strcmp(shpname, "CYCL") ) {

                    /* Make the walls structures instead */
                    p::uspool->createStructure(shpname, linenum, p::ppool->getPlayerNum("neutral"), 256,0,false);
                } else if( !strncmp(shpname, "TI", 2) ) {
                    /* tiberium */
                    sscanf(shpname,"TI%i",&i);
                    overlaymatrix[linenum] |= (i-1);
                } else {
                    strcat(shpname, ".SHP");

                    /* search the map for the image */
                    imgpos = imagelist.find(shpname);

                    /* set up the overlay matrix and load some shp's */
                    if( imgpos != imagelist.end() ) {

                        /* this tile already has a number */
                        overlaymatrix[linenum] |= HAS_OVERLAY;
                        overlays[linenum] = imgpos->second;
                    } else {

                        /* a new tile */
                        imagelist[shpname] = pc::imagepool->size();
                        overlaymatrix[linenum] |= HAS_OVERLAY;
                        overlays[linenum] = pc::imagepool->size();
                        try {
                            image = new SHPImage(shpname, mapscaleq);
                        } catch(ImageNotFound&) {
                            shpname[strlen(shpname)-3] = '\0';
                            strncat(shpname, missionData.theater, 3);
                            image = new SHPImage(shpname, mapscaleq);
                        }
                        pc::imagepool->push_back(image);
                    }

                    if( shpname[0] == 't' || shpname[0] == 'T' )
                        terraintypes[linenum] = t_tree;
                    else if( shpname[0] == 'r' || shpname[0] == 'R' )
                        terraintypes[linenum] = t_rock;
                    else
                        terraintypes[linenum] = t_other_nonpass;
                }

            }
        }
    } catch(int) {}

    try {
        for( keynum = 0;;keynum++ ) {
            key = inifile->readKeyValue("SMUDGE", keynum);
            /* , is the char which separate terraintype from action. */
            if( sscanf(key->first.c_str(), "%d", &linenum) == 1 &&
                    sscanf(key->second.c_str(), "SC%d", &smudgenum) == 1 ) {
                translateCoord(linenum, &tx, &ty);
                if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                    continue;
                }
                linenum = (ty-y)*width + tx - x;
                overlaymatrix[linenum] |= (smudgenum<<4);
            } else if( sscanf(key->first.c_str(), "%d", &linenum) == 1 &&
                       sscanf(key->second.c_str(), "CR%d", &smudgenum) == 1 ) {
                //} else if( sscanf(line, "%d=CR%d", &linenum, &smudgenum) == 2 ) {
                translateCoord(linenum, &tx, &ty);
                if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                    continue;
                }

                linenum = (ty-y)*width + tx - x;
                overlaymatrix[linenum] |= ((smudgenum+6)<<4);
            }
        }
    } catch(int) {}

    try {
        for( keynum = 0;;keynum++ ) {
            key = inifile->readKeyValue("STRUCTURES", keynum);
            /* , is the char which separate terraintype from action. */
            if( sscanf(key->first.c_str(), "%d", &tmpval) == 1 &&
                    sscanf(key->second.c_str(), "%[^,],%[^,],%d,%d,%d,%s", owner, type,
                           &health, &linenum, &facing, trigger ) == 6  ) {
                translateCoord(linenum, &tx, &ty);
                facing = min(31,facing>>3);
                if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                    continue;
                }
                linenum = (ty-y)*width + tx - x;
                p::uspool->createStructure(type, linenum, p::ppool->getPlayerNum(owner),
                        health, facing, false);
            }
        }
    } catch(int) {}

    try {
        for( keynum = 0;;keynum++ ) {
            key = inifile->readKeyValue("UNITS", keynum);
            /* , is the char which separate terraintype from action. */
            if( sscanf(key->first.c_str(), "%d", &tmpval) == 1 &&
                    sscanf(key->second.c_str(), "%[^,],%[^,],%d,%d,%d,%[^,],%s", owner, type,
                           &health, &linenum, &facing, action, trigger ) == 7  ) {
                translateCoord(linenum, &tx, &ty);
                facing = min(31,facing>>3);
                if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                    continue;
                }
                linenum = (ty-y)*width + tx - x;
                p::uspool->createUnit(type, linenum, 5, p::ppool->getPlayerNum(owner),
                        health, facing);
            }
        }
    } catch(int) {}

    /*infantry*/
    try {
        for( keynum = 0;;keynum++ ) {
            key = inifile->readKeyValue("INFANTRY", keynum);
            /* , is the char which separate terraintype from action. */
            if( sscanf(key->first.c_str(), "%d", &tmpval ) == 1  &&
                    sscanf(key->second.c_str(), "%[^,],%[^,],%d,%d,%d,%[^,],%d,%s", owner, type,
                           &health, &linenum, &subpos, action, &facing, trigger ) == 8  ) {
                translateCoord(linenum, &tx, &ty);
                facing = min(31,facing>>3);
                if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                    continue;
                }
                linenum = (ty-y)*width + tx - x;
                p::uspool->createUnit(type, linenum, subpos, p::ppool->getPlayerNum(owner),
                        health, facing);
            }
        }
    } catch(int) {}
    delete arts;
}



/** Load the bin part of the map. This contains the tiles.
 */

struct tiledata
{
    Uint32 image;
    Uint8 type;
};

void CnCMap::loadBin()
{
    Uint32 index = 0;
    //    Uint8 templ, tile;
    int xtile, ytile;
    VFile *binfile;
    char *binname = new char[strlen(missionData.mapname)+5];

    TileList *mapdata;

    mapdata = new TileList[width*height];

    /* Calculate name of bin file ( mapname.bin ). */
    strcpy(binname, missionData.mapname);
    strcat(binname, ".BIN");

    /* get the offset and size of the binfile along with a pointer to it */
    //binfile = mixes->getOffsetAndSize(binname, &offset, &size);
    binfile = VFS_Open(binname);
    delete[] binname;

    if(binfile == NULL) {
        logger->error("Unable to locate BIN file!\n");
        throw LoadMapError();
    }

    /* Seek the beginning of the map.
     * It's at begining of bin + maxwidth * empty y cells + empty x cells
     * times 2 sinse each entry is 2 bytes
     */
    binfile->seekSet( (64*y + x) * 2 );

    for( ytile = 0; ytile < height; ytile++ ) {
        for( xtile = 0; xtile < width; xtile++ ) {
            Uint16 tmpread = 0;
            /* Read template and tile */
            mapdata[index].templateNum = 0;
            binfile->readByte((Uint8 *)&(tmpread), 1);
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

            mapdata[index].templateNum = SDL_Swap16(tmpread);
#else

            mapdata[index].templateNum = tmpread;
#endif

            binfile->readByte(&(mapdata[index].tileNum), 1);

            index++;
        }
        /* Skip til the end of the line and the onwards to the
         * beginning of usefull data on the next line 
         */
        binfile->seekCur( 2*(64-width) );
    }
    VFS_Close(binfile);
    parseBin(mapdata);
    delete[] mapdata;
}

void CnCMap::unMapPack(INIFile* inifile)
{
    int tmpval;
    Uint32 curpos;
    int xtile, ytile;
    TileList *bindata;
    Uint32 keynum;
    INIKey key;
    Uint8 *mapdata1 = new Uint8[49152]; // 48k
    Uint8 *mapdata2 = new Uint8[49152];

    // read packed data into array
    mapdata1[0] = 0;
    try {
        for (keynum = 1;;++keynum) {
            key = inifile->readIndexedKeyValue("MAPPACK", keynum);
            strcat(((char*)mapdata1), key->second.c_str());
        }
    } catch(int) {}

    Compression::dec_base64(mapdata1,mapdata2,strlen(((char*)mapdata1)));

    /* decode the format80 coded data (6 chunks) */
    curpos = 0;
    for( tmpval = 0; tmpval < 6; tmpval++ ) {
        //printf("first vals in data is %x %x %x %x\n", mapdata2[curpos],
        //mapdata2[curpos+1], mapdata2[curpos+2], mapdata2[curpos+3]);
        if( Compression::decode80((Uint8 *)mapdata2+4+curpos, mapdata1+8192*tmpval) != 8192 ) {
            logger->warning("A format80 chunk in the \"MapPack\" was of wrong size\n");
        }
        curpos = curpos + 4 + mapdata2[curpos] + (mapdata2[curpos+1]<<8) +
                 (mapdata2[curpos+2]<<16);
    }
    delete[] mapdata2;

    /* 128*128 16-bit template number followed by 128*128 8-bit tile numbers */
    bindata = new TileList[width*height];
    tmpval = y*128+x;
    curpos = 0;
    for( ytile = 0; ytile < height; ytile++ ) {
        for( xtile = 0; xtile < width; xtile++ ) {
            /* Read template and tile */
            bindata[curpos].templateNum = ((Uint16 *)mapdata1)[tmpval];
            bindata[curpos].tileNum = mapdata1[tmpval+128*128*2];
            curpos++;
            tmpval++;
            /*  printf("tile %d, %d\n", bindata[curpos-1].templateNum,
                 bindata[curpos-1].tileNum); */
        }
        /* Skip til the end of the line and the onwards to the
         * beginning of usefull data on the next line 
         */
        tmpval += (128-width);
    }
    delete[] mapdata1;
    parseBin(bindata);
    delete[] bindata;
}

void CnCMap::parseBin(TileList* bindata)
{
    Uint32 index;

    Uint16 templ;
    Uint8 tile;
    int xtile, ytile;
    INIFile *templini;

    SDL_Surface *tileimg;
    SDL_Color palette[256];

    std::map<Uint32, struct tiledata> tilelist;
    std::map<Uint32, struct tiledata>::iterator imgpos;
    //     char tempc[sizeof(Uint8)];

    struct tiledata tiledata;
    Uint32 tiletype;
    tilematrix = new Uint16[width*height];

    loadPal(palette);
    SHPBase::setPalette(palette);
    SHPBase::calculatePalettes();

    /* Load the templates.ini */
    try {
        if( maptype == GAME_RA ) {
            templini = new INIFile("ratemplates.ini");
        } else {
            templini = new INIFile("templates.ini");
        }
    } catch(int) {
        logger->error("Could not load the templates file\n");
        throw LoadMapError();
    }
    index = 0;
    for( ytile = 0; ytile < height; ytile++ ) {
        for( xtile = 0; xtile < width; xtile++ ) {
            /* Read template and tile */
            templ = bindata[index].templateNum;
            tile  = bindata[index].tileNum;
            index++;
            /* Template 0xff is an empty tile */
            if(templ == ((maptype == GAME_RA)?0xffff:0xff)) {
                templ = 0;
                tile = 0;
            }

            /* Code sugested by Olaf van der Spek to cause all tiles in template
             * 0 and 2 to be used */
            if( templ == 0)
                tile = xtile&3 | (ytile&3 << 2);
            else if( templ == 2 )
                tile = xtile&1 | (ytile&1 << 1);


            imgpos = tilelist.find(templ<<8 | tile);

            /* set up the tile matrix and load some tiles */
            if( imgpos != tilelist.end() ) {
                /* this tile already has a number */
                tilematrix[width*ytile+xtile] = imgpos->second.image;
                if( terraintypes[width*ytile+xtile] == 0 )
                    terraintypes[width*ytile+xtile] = imgpos->second.type;
            } else {
                /* a new tile */
                tileimg = loadTile(templini, templ, tile, tiletype);

                tiledata.image = tileimages.size();
                tiledata.type = tiletype;
                tilelist[templ<<8 | tile] = tiledata;
                tilematrix[width*ytile+xtile] = tileimages.size();
                if( terraintypes[width*ytile+xtile] == 0 )
                    terraintypes[width*ytile+xtile] = tiletype;

                if( tileimg == NULL ) {
                    logger->error("Error loading tiles\n");
                    throw LoadMapError();
                }

                tileimages.push_back(tileimg);
            }

        }
    }
    delete templini;
}

/** Load a palette
 * @param the mixfiles.
 * @param a structure to store the palette in.
  */

void CnCMap::loadPal( SDL_Color *palette )
{
    VFile *mixfile;
    int i;
    char *palname = new char[strlen(missionData.theater)+5];

    /* Calculate the name of the palette ( theater.pal ) */
    strncpy(palname, missionData.theater, 8);
    palname[8] = '\0';
    strcat(palname, ".PAL");

    /* Seek the palette file in the mix */
    mixfile = VFS_Open(palname);

    if(mixfile == NULL) {
        logger->error("Unable to locate palette! This shouldn't happen unles you do something stupid\n"
                      "name: %s\n",palname);
        // abort();
        return;
    }

    delete[] palname;


    /* Load the palette */
    for (i = 0; i < 256; i++) {
        mixfile->readByte(&palette[i].r, 1);
        mixfile->readByte(&palette[i].g, 1);
        mixfile->readByte(&palette[i].b, 1);
        palette[i].r <<= 2;
        palette[i].g <<= 2;
        palette[i].b <<= 2;
    }
    VFS_Close(mixfile);
}



/** load a tile from the mixfile.
 * @param the mixfiles.
 * @param the template inifile.
 * @param the template number.
 * @param the tilenumber.
 * @returns a SDL_Surface containing the tile.
 */
SDL_Surface *CnCMap::loadTile( INIFile *templini,
                               Uint16 templ, Uint8 tile, Uint32 &tiletype )
{

    TemplateImage *theaterfile;

    SDL_Surface *retimage;


    char tilefilename[13];
    char tilenum[11];
    char *temname;


    /* The name of the file containing the template is something from
     * templates.ini . the three first
     * chars in the name of the theater eg. .DES .TEM .WIN */

    sprintf(tilefilename, "TEM%d", templ);

    sprintf(tilenum, "tiletype%d", tile);
    tiletype = templini->readInt(tilefilename, tilenum);
    if( tiletype == INIERROR )
        tiletype = 0;
    //   printf("Type is %d\n", tiletype);

    temname = templini->readString(tilefilename, "Name");

    if( temname == NULL ) {
        logger->warning("Error in templates.ini! (can't find \"%s\")\n", tilefilename);
        strcpy(tilefilename, "CLEAR1");
    } else {
        strcpy(tilefilename, temname);
        delete[] temname;
    }

    strcat( tilefilename, "." );
    strncat( tilefilename, missionData.theater, 3 );


    /* Find and seek the template file in the mix */

    try {
        if( maptype == GAME_RA ) {
            theaterfile = new TemplateImage(tilefilename, mapscaleq, 1);
        } else {
            theaterfile = new TemplateImage(tilefilename, mapscaleq);
        }
    } catch(ImageNotFound&) {
        logger->warning("Unable to locate template %d, %d (\"%s\") in mix! using tile 0, 0 instead\n", templ, tile, tilefilename);
        if( templ == 0 && tile == 0 )
            return NULL;
        return loadTile( templini, 0, 0, tiletype );
    }


    retimage = theaterfile->getImage(tile);
    if( retimage == NULL ) {
        logger->warning("Illegal template %d, %d (\"%s\")! using tile 0, 0 instead\n", templ, tile, tilefilename);
        if( templ == 0 && tile == 0 )
            return NULL;
        return loadTile( templini, 0, 0, tiletype );
    }
    delete theaterfile;
    return retimage;

}

