#include <algorithm>
#include <cstdio>
#include <cstdarg>
#include <cctype>
#include <string>
#include "SDL_endian.h"
#include "externalvfs.h"
#include "common.h"

using std::string;

static FILE* fcaseopen(const char* path, const char* mode) {
    FILE* ret;
    string fname = path;
    ret = fopen(fname.c_str(), mode);
    if (ret != NULL) {
        return ret;
    }
    // Try all lower
    transform(fname.begin(), fname.end(), fname.begin(), tolower);
    ret = fopen(fname.c_str(), mode);
    if (ret != NULL) {
        return ret;
    }
    /// @TODO Try other tricks like "lower.EXT" or "UPPER.ext"
    return NULL;
}

ExternalFiles::ExternalFiles(const char *defpath)
{
    this->defpath = cppstrdup(defpath);
}

ExternalFiles::~ExternalFiles()
{
    unloadArchives();
    delete[] defpath;
}

bool ExternalFiles::loadArchive(const char *fname)
{
    if( isRelativePath(fname) ) {
        char *pth;
        pth = new char[strlen(defpath)+strlen(fname)+2];
        sprintf(pth, "%s/%s", defpath, fname);
        path.push_back(pth);
        return true;
    }
    path.push_back(cppstrdup(fname));
    return true;
}

void ExternalFiles::unloadArchives()
{
    Uint32 i;

    for( i = 0; i < path.size(); i++ ) {
        delete[] path[i];
    }
    path.resize(0);
}


Uint32 ExternalFiles::getFile(const char *fname, const char* mode)
{
    openFileExt newFile;
    FILE *f;
    Uint32 i;
    char filename[1024];
    Uint32 size, fnum;

    if (mode[0] != 'r') {
        memset(filename,0,1024);
        strcpy(filename, fname);
        f = fopen(filename, mode);
        if (f != NULL) {
            newFile.file = f;
            /* We'll just ignore file sizes for files being written for now.
             */
            newFile.size = 0;
            newFile.path = cppstrdup(filename);
            fnum = (Uint32)f;
            openFiles[fnum] = newFile;
            return fnum;
        }
    }
    for( i = 0; i < path.size(); i++ ) {
        strcpy(filename, path[i]);
        strcat(filename, fname);
        f = fcaseopen(filename, mode);
        if( f != NULL ) {
            fseek(f, 0, SEEK_END);
            size = ftell(f);
            fseek(f, 0, SEEK_SET);
            newFile.file = f;
            newFile.size = size;

            newFile.path = cppstrdup(filename);

            fnum = (Uint32)f;
            openFiles[fnum] = newFile;
            return fnum;
        }
    }

    return (Uint32)-1;
}

void ExternalFiles::releaseFile(Uint32 file)
{
    fclose(openFiles[file].file);
    delete[] openFiles[file].path;
    openFiles.erase(file);
}

Uint32 ExternalFiles::readByte(Uint32 file, Uint8 *databuf, Uint32 numBytes)
{
    return fread(databuf, 1, numBytes, openFiles[file].file);
}

Uint32 ExternalFiles::readWord(Uint32 file, Uint16 *databuf, Uint32 numWords)
{
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = fread(databuf, 2, numWords, openFiles[file].file);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        databuf[i] = SDL_Swap16(databuf[i]);
    }
#endif

    return numRead;
}

Uint32 ExternalFiles::readThree(Uint32 file, Uint32 *databuf, Uint32 numThrees)
{
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = fread(databuf, 3, numThrees, openFiles[file].file);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        databuf[i] = SDL_Swap32(databuf[i]);
        databuf[i]<<=8;
    }
#endif

    return numRead;
}

Uint32 ExternalFiles::readDWord(Uint32 file, Uint32 *databuf, Uint32 numDWords)
{
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = fread(databuf, 4, numDWords, openFiles[file].file);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        databuf[i] = SDL_Swap32(databuf[i]);
    }
#endif

    return numRead;
}

char *ExternalFiles::readLine(Uint32 file, char *databuf, Uint32 buflen)
{
    return fgets(databuf, buflen, openFiles[file].file);
}

Uint32 ExternalFiles::writeByte(Uint32 file, const Uint8* databuf, Uint32 numBytes)
{
    return fwrite(databuf, 1, numBytes, openFiles[file].file);
}

Uint32 ExternalFiles::writeWord(Uint32 file, const Uint16 *databuf, Uint32 numWords)
{
    Uint32 numWrote;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint16* tmp = new Uint16[numWords];
    Uint32 i;

    for( i = 0; i < numWords; i++ ) {
        tmp[i] = SDL_Swap16(databuf[i]);
    }

    numWrote = fwrite(tmp, 2, numWords, openFiles[file].file);
    delete[] tmp;
#else

    numWrote = fwrite(databuf, 2, numWords, openFiles[file].file);
#endif

    return numWrote;
}

Uint32 ExternalFiles::writeThree(Uint32 file, const Uint32 *databuf, Uint32 numThrees)
{
    Uint32 numWrote;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32* tmp = new Uint32[numThrees];
    Uint32 i;

    for( i = 0; i < numThrees; i++ ) {
        tmp[i] = SDL_Swap32(databuf[i]);
        tmp[i]<<=8;
    }
    numWrote = fwrite(tmp, 3, numThrees, openFiles[file].file);
    delete[] tmp;
#else

    numWrote = fwrite(databuf, 3, numThrees, openFiles[file].file);
#endif

    return numWrote;
}

Uint32 ExternalFiles::writeDWord(Uint32 file, const Uint32 *databuf, Uint32 numDWords)
{
    Uint32 numWrote;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
    Uint32* tmp = new Uint32[numDWords];

    for( i = 0; i < numDWords; i++ ) {
        tmp[i] = SDL_Swap32(databuf[i]);
    }
    numWrote = fwrite(tmp, 4, numDWords, openFiles[file].file);
    delete[] tmp;
#else

    numWrote = fwrite(databuf, 4, numDWords, openFiles[file].file);
#endif

    return numWrote;
}

void ExternalFiles::writeLine(Uint32 file, const char *databuf)
{
    fputs(databuf, openFiles[file].file);
}

int ExternalFiles::vfs_printf(Uint32 file, const char* fmt, va_list ap)
{
    int ret;
    ret = vfprintf(openFiles[file].file, fmt, ap);
    return ret;
}

void ExternalFiles::seekSet(Uint32 file, Uint32 pos)
{
    fseek(openFiles[file].file, pos, SEEK_SET);
}

void ExternalFiles::seekCur(Uint32 file, Sint32 pos)
{
    fseek(openFiles[file].file, pos, SEEK_CUR);
}

Uint32 ExternalFiles::getPos(Uint32 file)
{
    return ftell(openFiles[file].file);
}

Uint32 ExternalFiles::getSize(Uint32 file)
{
    return openFiles[file].size;
}

char *ExternalFiles::getPath(Uint32 file)
{
    return openFiles[file].path;
}


